from rich.console import Console
from langchain_ollama.llms import OllamaLLM
from langchain_core.messages import HumanMessage, AIMessage
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder

from .base_agent import BaseAgent

SYSTEM_TEMPLATE = """\
You are an exeprt in answering questions about reseach papers.
If you don't know the answer, just say that you don't know.
"""

HUMAN_TEMPLATE = """\
Use the following relevant context to answer the question.
Make sure to cite the source documents if you can.

----------------Context:
{context}

----------------Question:
{question}
"""


class LocalDocAgent(BaseAgent):
    def __init__(self, text_embedder):
        self.template = SYSTEM_TEMPLATE

        self.model = OllamaLLM(model="llama3.2")
        self.prompt = ChatPromptTemplate([
            ("system", SYSTEM_TEMPLATE),
            MessagesPlaceholder(variable_name="chat_history"),
            ("human", HUMAN_TEMPLATE),
        ])
        self.chain = self.prompt | self.model

        self.chat_history = []

        self.text_embedder = text_embedder

    def retrieve(self, query: str) -> str:
        with Console().status('', spinner='dots'):
            context = self.text_embedder.invoke(query)
            response = self.chain.invoke({
                "context": context,
                "question": query,
                "chat_history": self.chat_history,
            })
            self.chat_history.extend([
                HumanMessage(content=query),
                AIMessage(content=response),
            ])
            return response



import os
from typing import Optional
from dataclasses import dataclass
from rich.console import Console
from pydantic import BaseModel, Field
from pydantic_ai import Agent, RunContext
from pydantic_ai.models.openai import OpenAIChatModel
from pydantic_ai.providers.openai import OpenAIProvider
from langchain_core.vectorstores import VectorStoreRetriever

from .base_agent import BaseAgent

DEFAULT_OLLAMA_URL = os.getenv("OLLAMA_HOST", "http://localhost:11434/v1")

AGENT_TEMPLATE = """\
You are an exeprt in answering questions about reseach papers.
Use the following relevant context to answer the question.
If you don't know the answer, just say that you don't know.
Make sure to cite the source documents if you can.
If you are confident, generate the final answer without repeating tool use.
When you do not need to use a tool, answer directly.
Do not review the same document more than once.
"""


class RAGAnswer(BaseModel):
    """Structured response for a RAG question."""
    response: str = Field(
        description="The final answer based on the retrieved context.")


@dataclass
class Deps:
    retriever: VectorStoreRetriever


class PydanticAIDocAgent(BaseAgent):
    def __init__(self, text_embedder) -> None:
        ollama_model = OpenAIChatModel(
            model_name='llama3.2',
            provider=OpenAIProvider(base_url=DEFAULT_OLLAMA_URL),
        )
        self.agent = Agent(
            ollama_model,
            system_prompt=AGENT_TEMPLATE,
            deps_type=Deps,
            output_type=str,
        )

        self.deps = Deps(text_embedder.invoke)

        @self.agent.tool_plain
        def list_documents_titles() -> str:
            """
            List the titles of all the documents provided by the user.
            """
            collection = text_embedder.vector_db._collection
            metadatas = collection.get(include=["metadatas"])["metadatas"]

            unique_titles = {meta["title"] for meta in metadatas
                             if meta and "title" in meta}
            response = "\n\n---\n\n".join(
                f"{i+1}. {title}" for i, title in enumerate(unique_titles)
            )

            return response

        @self.agent.tool
        def retrieve_context(ctx: RunContext[Deps], query: str,
                             file_index: Optional[int] = None) -> str:
            """
            Searches and retrieves relevant document to answer user
            questions about research papers.
            Use this tool to find factual information, definitions,
            or specific details.
            If the user asks about a specific file, defined by its index
            `file_index`, apply a filter to invoke.
            If the information is to be retrieved from all the documents,
            set `file_index` to None.
            """
            kwargs = dict()
            if file_index is not None:
                kwargs['filter'] = {'file_index': file_index}
            c = f"Context:\n{ctx.deps.retriever(query, **kwargs)}"
            # c = f"Context:\n{text_embedder.invoke(query, **kwargs)}"
            return c

        self.message_history = None

    def retrieve(self, query: str) -> str:
        with Console().status('', spinner='dots'):
            result = self.agent.run_sync(
                query, message_history=self.message_history,
                deps=self.deps,
            )
            self.message_history = result.new_messages()
            return result.output



import os
import shutil
import requests
import pymupdf
from typing import Iterable
from concurrent.futures import ThreadPoolExecutor, as_completed
from rich.progress import (Progress, SpinnerColumn, BarColumn,
                           MofNCompleteColumn, TextColumn)
from pathlib import Path
from tempfile import NamedTemporaryFile
from docling.document_converter import DocumentConverter
from docling.chunking import HybridChunker
from langchain_ollama import OllamaEmbeddings
from langchain_chroma import Chroma
from langchain_core.documents import Document
from langchain_text_splitters import RecursiveCharacterTextSplitter
from flashrank import Ranker
from langchain_community.document_compressors import FlashrankRerank
from langchain.retrievers import ContextualCompressionRetriever

from .utils import get_sitemap_urls

import warnings
from transformers import logging

logging.set_verbosity_error()
warnings.filterwarnings("ignore")

# default value for max_workers
MAX_WORKERS = os.cpu_count() or 4


def download_url(url: str) -> str:
    try:
        response = requests.get(url)
        response.raise_for_status()
    except Exception as e:
        print(e)
        return None

    tmp_file = NamedTemporaryFile(delete=False)
    tmp_file.write(response.content)
    tmp_file.flush()
    return tmp_file.name


def download_all(urls: list[str], max_workers=MAX_WORKERS):
    results = []
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        BarColumn(),
        MofNCompleteColumn(),
        TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
    ) as progress:
        task_progress = progress.add_task(
            "[cyan]Downloading urls ...", total=len(urls))

        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            futures = [executor.submit(download_url, url) for url in urls]

            for future in as_completed(futures):
                if (f_path := future.result()) is not None:
                    results.append(f_path)
                progress.update(task_progress, advance=1)

    return results


class TextEmbedderDB:
    def __init__(self, source: list[str | os.PathLike[str]],
                 *, url: Iterable[str] | None = None,
                 fname: Iterable[str | os.PathLike[str]] | None = None,
                 path_db: str | os.PathLike[str] | None = None,
                 topk: int = 5, use_reranker: bool = True) -> None:
        if not (source is not None and len(source) > 0):
            raise ValueError('source should not be None or an empty list')

        self.source = source
        self.topk = topk
        self.use_reranker = use_reranker
        self.model_embeddings = OllamaEmbeddings(model="mxbai-embed-large")

        self.init_db(path_db)
        self.prepare_source()
        self.process_documents()
        self.init_retriever()

    def prepare_source(self):
        sitemap_urls = []

        for (i, source) in enumerate(self.source[:]):
            if str(source).endswith('/sitemap.xml'):
                self.source.pop(i)
                try:
                    urls = get_sitemap_urls(source)
                    sitemap_urls.extend(urls)
                except Exception:
                    print("Could not extract urls from:", source)

        if len(sitemap_urls) > 0:
            print('Found URLs in the sitemaps.')
        self.source.extend(sitemap_urls)

    def _process_document(self, index: int,
                          source: str | os.PathLike[str]) -> None:
        document = DocumentConverter().convert(source).document
        chunks = HybridChunker().chunk(document)

        ids = []
        documents = []
        title = None
        for i, chunk in enumerate(chunks):
            id = f"{index}-{i}"
            ids.append(id)
            title, doc = self._format_chunk(id, index, chunk, title)
            documents.append(doc)

        self.vector_db.add_documents(ids=ids, documents=documents)

    def _format_chunk(self, id: str, index: int,
                      chunk, title: str | None):
        heading = (chunk.meta.headings
                   if hasattr(chunk.meta, 'headings')
                   else "Unknown Heading")
        if isinstance(heading, Iterable):
            heading = ", ".join(heading)
        if title is None:
            # NOTE: heuristic -> setting the title to be the 1st heading
            title = heading

        page_numbers = ', '.join(sorted(list(set(
            str(p.page_no) for item in chunk.meta.doc_items for p in item.prov
        ))))

        metadata = {
            "title": title,
            "heading": heading,
            "filename": chunk.meta.origin.filename,
            "page_numbers": page_numbers,
            "file_index": index,
        }
        doc = Document(page_content=chunk.text, metadata=metadata, id=id)
        return title, doc

    def process_documents(self) -> None:
        self.ids = []
        self.documents = []
        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            MofNCompleteColumn(),
            TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
        ) as progress:
            task_progress = progress.add_task(
                "[cyan]Processing the documents ...",
                total=len(self.source))

            with ThreadPoolExecutor(max_workers=MAX_WORKERS) as executor:
                futures = {
                    executor.submit(self._process_document, i, s): i
                    for (i, s) in enumerate(self.source)
                }

                for _ in as_completed(futures):
                    # ids, documents = future.result()
                    progress.update(task_progress, advance=1)

    def add_new_document(self, url: str | None = None,
                         fname: str | os.PathLike[str] | None = None):
        if not ((url is not None) ^ (fname is not None)):
            raise ValueError('Either `url` or `fpath` should be specified:',
                             f'{url=} - {fname=}')
        _source = fname or url
        self.source.append(_source)
        self._process_document(len(self.source), _source)

    def init_retriever(self):
        base_retriever = self.vector_db.as_retriever(
            search_kwargs={"k": self.topk})

        if self.use_reranker:
            compressor = FlashrankRerank(
                client=Ranker(model_name="ms-marco-MultiBERT-L-12",
                              log_level='WARNING'),
                top_n=self.topk // 2)
            self.retriever = ContextualCompressionRetriever(
                base_compressor=compressor, base_retriever=base_retriever)
        else:
            self.retriever = base_retriever

    def invoke(self, query: str, **kwargs) -> str:
        retrieved_docs = self.retriever.invoke(query, **kwargs)

        formatted_docs = []
        for i, doc in enumerate(retrieved_docs, 1):
            metadata = doc.metadata
            title = metadata.get("title", "Unknown Title")
            file_index = metadata.get("file_index", "N/A")
            filename = metadata.get("filename", "Unknown filename")
            heading = metadata.get("heading", "Unknown Heading")
            page_numbers = metadata.get('page_numbers', 'N/A')

            formatted_docs.append(
                f"### Chunk ID {doc.id}\n"
                f" **Document Title:** {title}\n"
                f" **Document index:** {file_index}\n"
                f" **Document filename:** {filename}\n"
                f" **Heading:** {heading}\n"
                f" **Page(s): {page_numbers}\n"
                f"---\n{doc.page_content}\n"
            )
        return "\n\n".join(formatted_docs)

    def init_db(self, path_db: str | os.PathLike[str] | None) -> None:
        default_path = (Path(__file__).resolve().absolute().parents[2]
                        / 'vector_db')

        path_db = path_db or default_path
        path_db = Path(path_db).resolve().absolute()

        self.path_db = path_db if path_db.exists() else default_path
        shutil.rmtree(self.path_db, ignore_errors=True)

        self.vector_db = Chroma(
            collection_name='vector_db',
            embedding_function=self.model_embeddings,
            persist_directory=self.path_db,
        )


class TextExtractor:
    def __init__(self, *, url: str | None = None,
                 fname: str | os.PathLike[str] | None = None,
                 chunk_size: int = 100, chunk_overlap: int = 100) -> None:
        # XOR operator
        if not ((url is not None) ^ (fname is not None)):
            raise ValueError('Either `url` or `fname` should be specified:',
                             f'{url=} - {fname=}')

        self.text_splitter = RecursiveCharacterTextSplitter(
            chunk_size=chunk_size, chunk_overlap=chunk_overlap,
            length_function=len, is_separator_regex=False,
        )

        self._text = self.process(url=url, fname=fname)

    @property
    def text(self) -> str:
        return self._text

    @property
    def chunks(self) -> list[str]:
        return self.text_splitter.split_text(self._text)

    def process(self, *, url: str | None = None,
                fname: str | os.PathLike[str] | None = None) -> str:
        # save the pdf in the url to a temporary file
        if url is not None:
            response = requests.get(url)
            response.raise_for_status()

            tmp_file = NamedTemporaryFile(delete=False)
            tmp_file.write(response.content)
            tmp_file.flush()
            fname = tmp_file.name

        assert fname is not None
        text = self.extract_text(fname)

        # delete the temporary file
        if url is not None:
            os.remove(fname)

        return text

    def extract_text(self, fname: str | os.PathLike[str] | None) -> str:
        text = ""

        try:
            with pymupdf.open(fname) as doc:  # type: ignore[no-untyped-call]
                for page in doc:
                    text += page.get_text()
        except Exception as e:
            print(e)

        return text



import torch
import numpy as np
from transformers import AutoTokenizer, AutoModelForCausalLM


class Encoder:
    model_name = "google/gemma-3-270m"

    def __init__(self):
        self.tokenizer = AutoTokenizer.from_pretrained(self.model_name)
        self.model = AutoModelForCausalLM.from_pretrained(self.model_name)
        self.embedder = self.get_embeddings()

    def get_embeddings(self):
        @torch.no_grad()
        def _embeddings(text):
            tokens = self.tokenizer(text)['input_ids']
            if isinstance(tokens[0], int):
                tokens = [tokens]

            return [
                self.model.model.embed_tokens(
                    torch.tensor(t, dtype=torch.int)
                ).detach().cpu().numpy()
                for t in tokens
            ]
        return _embeddings

    def encode(self, text: str | list[str], **kwargs) -> np.ndarray:
        if isinstance(text, str):
            text = [text]

        # Tokenize with padding and truncation
        inputs = self.tokenizer(
            text,
            return_tensors="pt",
            padding=True,
            truncation=True
        )

        outputs = self.model(**inputs, output_hidden_states=True,
                             return_dict=True)
        hidden_states = outputs.hidden_states[-1]

        # Apply mean pooling (ignoring padded tokens)
        attention_mask = inputs["attention_mask"]
        mask = attention_mask.unsqueeze(-1).expand(
            hidden_states.size()).float()
        masked_embeddings = hidden_states * mask

        sum_embeddings = torch.sum(masked_embeddings, dim=1)
        sum_mask = torch.clamp(mask.sum(dim=1), min=1e-9)
        return (sum_embeddings / sum_mask).detach().cpu().numpy()



import os
import sys
import shutil
import argparse

from . import agents as my_agents
from .formatter import TerminalIO
from .processing import TextEmbedderDB


def answer_one_query(agent: my_agents.BaseAgent,
                     console: TerminalIO) -> None:
    try:
        query = console.ask()
        if query == "clear":
            os.system('cls' if os.name == 'nt' else 'clear')
            return
    except (KeyboardInterrupt, EOFError):
        res = input("\nDo you really want to exit ([y]/n)? ").lower()
        if res in ("", "y", "yes"):
            console.answer("Bye Bye!")
            sys.exit()
        else:
            return
    console.answer(agent.retrieve(query))


def main() -> None:
    parser = argparse.ArgumentParser('DocSeer')
    parser.add_argument(
        '-u', '--url', type=str, nargs='*', default=[],
    )
    parser.add_argument(
        '-f', '--file-path', type=str, nargs='*', default=[],
    )
    parser.add_argument(
        '-s', '--source', type=str, nargs='*', default=[],
    )
    parser.add_argument(
        '-a', '--arxiv-id', type=str, nargs='*', default=None,
    )
    parser.add_argument(
        '-k', '--top-k', type=int, default=10,
    )
    parser.add_argument(
        '-Q', '--query', type=str, default=None,
    )
    parser.add_argument(
        '-I', '--interactive', action='store_true',
    )
    args = parser.parse_args()

    if (args.query is None) and (not args.interactive):
        return

    # TODO: remove this line!
    args.source += args.file_path + args.url

    if args.arxiv_id is not None:
        args.source += [f"https://arxiv.org/pdf/{arxiv_id}"
                        for arxiv_id in args.arxiv_id]

    text_embedder = None
    try:
        console = TerminalIO(is_table=True)

        text_embedder = TextEmbedderDB(
            source=args.source, topk=args.top_k)

        agent = my_agents.LocalDocReActAgent(text_embedder)

        if args.interactive:
            while True:
                answer_one_query(agent, console)
        elif args.query is not None:
            console.answer(agent.retrieve(args.query))
    finally:
        # clean-ups: database
        path_db = getattr(text_embedder, 'path_db', None)
        if path_db is not None:
            shutil.rmtree(path_db, ignore_errors=True)
        ...


if __name__ == "__main__":
    main()



import torch
import numpy as np
from operator import itemgetter
from tqdm import tqdm
from faiss import IndexFlatIP, normalize_L2
from transformers import pipeline
from sentence_transformers import SentenceTransformer

from .base_agent import BaseAgent


class HFDocAgent(BaseAgent):
    sum_model = "Falconsai/text_summarization"
    enc_model = "sentence-transformers/all-MiniLM-L6-v2"
    qa_model = "distilbert/distilbert-base-cased-distilled-squad"

    def __init__(self, text: str, chunks: list[str],
                 device: torch.device = torch.device('cpu')) -> None:
        self.text = text
        self.chunks = chunks
        self.device = device

        self.encoder = SentenceTransformer(self.enc_model)
        self.question_answerer = pipeline("question-answering",
                                          model=self.qa_model,
                                          device=device)
        self.embeddings = self._encode(self.chunks)
        self.indexer = FAISSFindClosest(self.embeddings)

    @torch.no_grad()
    def _encode(self, text: str | list[str]) -> np.ndarray:
        return self.encoder.encode(text, convert_to_numpy=True)

    @torch.no_grad()
    def _summarizer(self, text: str) -> str:
        summarizer = pipeline("summarization", model=self.sum_model,
                              device=self.device)
        return summarizer(text, min_length=10, max_length=40,
                          max_new_tokens=None, do_sample=True,
                          truncation=True)[0]['summary_text']

    @torch.no_grad()
    def summarize(self) -> str:
        return " ".join(list(tqdm(map(self._summarizer, self.chunks),
                                  total=len(self.chunks))))

    @torch.no_grad()
    def retrieve(self, query: str) -> str:
        query_embedding = self._encode(query).reshape(1, -1)
        indices = self.indexer(query_embedding)
        closest_chunks = itemgetter(*indices)(self.chunks)

        return self.question_answerer(
            question=query, context="\n".join(closest_chunks)
        )['answer']


class FAISSFindClosest:
    def __init__(self, embeddings: np.ndarray) -> None:
        if embeddings.ndim != 2:
            raise ValueError("Embeddings must be a 2D array"
                             " (num_vectors, vector_dim)")

        # Copy to avoid mutating original embeddings
        self.embeddings = embeddings.astype(np.float32, copy=True)
        normalize_L2(self.embeddings)

        self.index = IndexFlatIP(self.embeddings.shape[-1])
        self.index.add(self.embeddings)

    def __call__(
        self,
        input_embedding: np.ndarray | list[float],
        top_k: int = 3,
        return_scores: bool = False
    ) -> list[int] | tuple[list[int], list[float]]:
        vec = np.asarray(input_embedding, dtype=np.float32)

        if vec.ndim == 1:
            vec = vec.reshape(1, -1)
        elif vec.ndim != 2 or vec.shape[0] != 1:
            raise ValueError("Input embedding must be a 1D vector"
                             " or a 2D array with shape (1, dim)")

        normalize_L2(vec)

        scores, indices = self.index.search(vec, top_k)

        if return_scores:
            return indices[0].tolist(), scores[0].tolist()
        return indices[0].tolist()



import torch
import requests
import xml.etree.ElementTree as ET


def get_device(device: str | None = None) -> torch.device:
    if device is not None:
        return torch.device(device)
    elif torch.backends.mps.is_available():
        return torch.device('mps')
    elif torch.cuda.is_available():
        return torch.device('cuda')
    else:
        return torch.device('cpu')


def get_sitemap_urls(sitemap_url: str) -> list[str]:
    """Fetches and parses a sitemap XML file to extract URLs.

    Args:
        sitemap_url: The sitemap.xml URL of the website

    Returns:
        List of URLs found in the sitemap. If sitemap is not found,
        returns a list containing only the base URL.

    Raises:
        ValueError: If there's an error fetching or parsing the sitemap
    """
    try:
        assert sitemap_url.endswith('sitemap.xml')

        # Fetch sitemap URL
        response = requests.get(sitemap_url, timeout=10)
        response.raise_for_status()

        # Parse XML content
        root = ET.fromstring(response.content)

        # Handle different XML namespaces that sitemaps might use
        namespaces = (
            {"ns": root.tag.split("}")[0].strip("{")}
            if "}" in root.tag else ""
        )

        # Extract URLs using namespace if present
        it = (root.findall(".//ns:loc", namespaces)
              if namespaces else root.findall(".//loc"))

        return [elem.text for elem in it]

    except requests.RequestException as e:
        raise ValueError(f"Failed to fetch sitemap: {str(e)}")
    except ET.ParseError as e:
        raise ValueError(f"Failed to parse sitemap XML: {str(e)}")
    except Exception as e:
        raise ValueError(f"Unexpected error processing sitemap: {str(e)}")



from typing import Any
from rich.live import Live
from rich.panel import Panel
from rich.markdown import Markdown
from rich.console import Console, Group
from langchain_core.agents import AgentFinish
from langchain.callbacks.base import BaseCallbackHandler


class RichCallbackHandler(BaseCallbackHandler):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.panels = []
        self.live = None
        self.main_panel = Panel(Group(*self.panels),
                                title="Agent Executor Log")
        self.console = Console()
        self.live = None

    def _update_live_display(self) -> None:
        """
        Helper to update the live display with the current group of panels.
        """
        if self.live is not None:
            self.main_panel.renderable = Group(*self.panels)
            self.live.refresh()

    def on_chain_start(self, serialized, inputs, **kwargs) -> None:
        if "name" in kwargs:
            name = kwargs["name"]
        elif serialized:
            name = serialized.get(
                "name", serialized.get("id", ["<unknown>"])[-1])
        else:
            name = "<unknown>"
        self.panels.append(
            Panel(Markdown(f"Entering new chain: {name}"),
                  title="Chain Started", style="bold blue")
        )

        self.main_panel.renderable = Group(*self.panels)
        self.live = Live(self.main_panel,
                         console=self.console,
                         vertical_overflow="visible",
                         # refresh_per_second=5,
                         auto_refresh=False)
        self.live.start()
        self.live.refresh()

    def on_chain_end(self, outputs: dict[str, Any], **kwargs) -> None:
        self.panels.append(
            Panel(Markdown(f"Output: {outputs}"),
                  title="Chain Finish", style="bold blue")
        )
        self._update_live_display()
        if self.live:
            self.live.stop()
        self.panels = []

    def on_chain_error(self, error, **kwargs) -> None:
        self.panels.append(
            Panel(Markdown(str(error)), title="Chain Error", style="bold red")
        )
        self._update_live_display()

    def on_tool_start(self, serialized, input_str, **kwargs) -> None:
        self.panels.append(
            Panel(Markdown(input_str),
                  title=f"Tool Started: {serialized['name']}",
                  style="bold yellow")
        )
        self._update_live_display()

    def on_tool_end(self, output, **kwargs) -> None:
        self.panels.append(
            Panel(Markdown(output), title="Tool Ended", style="bold yellow")
        )
        self._update_live_display()

    def on_agent_action(self, action, **kwargs) -> None:
        self.panels.append(
            Panel(action.log, title="Agent Action", style="bold cyan")
        )
        self._update_live_display()

    def on_agent_finish(self, finish: AgentFinish, **kwargs: Any) -> None:
        self.panels.append(
            Panel(Markdown(finish.log), title="Agent Finished",
                  style="bold cyan")
        )
        self._update_live_display()



from rich.markdown import Markdown
from rich.console import Console
from rich.style import Style
from rich.prompt import Prompt
from rich.panel import Panel


class TerminalIO:
    console = Console()
    question_style = "[bold slate_blue1]"
    print_style = Style(color="sea_green2", bold=False)

    def __init__(self, is_table: bool = True,
                 width: int | None = None) -> None:
        self.input_msg = f"{self.question_style}>>> question"
        self.is_table = is_table
        self.width = width

    def ask(self) -> str:
        return Prompt.ask(self.input_msg, show_default=False)

    def answer(self, response: str) -> None:
        response = Markdown(response)
        if self.is_table:
            self.console.print(Panel(
                response, style=self.print_style,
                width=self.width, expand=self.width is None))
        else:
            self.console.print(response, style=self.print_style)


if __name__ == "__main__":
    ter = TerminalIO()
    txt = ter.ask()
    ter.answer(f'hello {txt}')



import contextlib
from rich.console import Console
from langchain_core.tools import tool
from langchain_ollama.llms import OllamaLLM
from langchain_core.prompts import PromptTemplate
from langchain.memory import ConversationBufferMemory
from langchain.agents import AgentExecutor, create_react_agent

from .base_agent import BaseAgent
from .callback_handlers import RichCallbackHandler


AGENT_TEMPLATE = """\
You are an expert in answering questions about research papers.
You have access to the following tool:

{tools}

To use a tool, you must follow this format:
```
Thought: Do I need to use a tool? Yes.
Action: the action to take, should be one of [{tool_names}]
Action Input: the input to the action
Observation: the result of the action
... (this Thought/Action/Observation can repeat
 until you have a final answer)
Thought: I have enough information to answer the question.
Final Answer: [your final answer here]
```

If you are confident, generate the final answer without repeating tool use.
When you do not need to use a tool, answer directly.
Do not review the same document more than once.

Begin!

Chat History:
{chat_history}
Question: {question}
{agent_scratchpad}
"""


class LocalDocReActAgent(BaseAgent):
    """
    An agentic RAG system for answering questions about research papers.

    This agent can intelligently decide when to retrieve context from
    a knowledge base before answering a user's question.
    """

    def __init__(self, text_embedder, verbose=True):
        self.template = AGENT_TEMPLATE

        @tool
        def retrieve_document_context(query: str,
                                      file_index: int | None = None) -> str:
            """
            Searches and retrieves relevant document to answer user
            questions about research papers.
            Use this tool to find factual information, definitions,
            or specific details.
            If the user asks about a specific file, defined by its index
            `file_index`, apply a filter to invoke.
            If the information is to be retrieved from all the documents,
            set `file_index` to None.
            """
            kwargs = dict()
            if file_index is not None:
                kwargs['filter'] = {'file_index': file_index}
            return text_embedder.invoke(query, **kwargs)

        @tool
        def number_of_unique_documents() -> int:
            """
            Count the number of unique documents provided by the user.
            """
            return len(text_embedder.source)

        @tool
        def list_documents_titles() -> str:
            """
            List the titles of all the documents provided by the user.
            """
            collection = text_embedder.vector_db._collection
            metadatas = collection.get(include=["metadatas"])["metadatas"]

            unique_titles = {meta["title"] for meta in metadatas
                             if meta and "title" in meta}
            response = "\n\n---\n\n".join(
                f"{i+1}. {title}" for i, title in enumerate(unique_titles)
            )

            return response

        @tool
        def check_history_relevance(new_input: str) -> str:
            """
            Analyzes the chat history to determine if it is relevant to
            the new user input.
            Use this tool to decide if you need to perform a new search.
            Returns 'yes' if the history is relevant, 'no' otherwise.
            """
            # Use a lightweight LLM to check relevance.
            relevance_llm = OllamaLLM(model="llama3.2")
            relevance_prompt = PromptTemplate.from_template(
                "Based on the chat history below, "
                "is the new user input about a related topic? "
                "Respond with 'yes' or 'no' only."
                "\n\nChat History:\n{history}\n\nNew Input:{input}"
            )
            # We use the raw buffer string for a simple, quick check.
            response = relevance_llm.invoke(relevance_prompt.format(
                history=self.memory.buffer,
                input=new_input
            ))
            return response.strip().lower()

        self.tools = [retrieve_document_context, check_history_relevance,
                      number_of_unique_documents, list_documents_titles]

        # Create the LangChain PromptTemplate and the LLM.
        self.prompt = PromptTemplate.from_template(self.template)
        self.llm = OllamaLLM(model="llama3.2")
        self.verbose = verbose

        self.agent = create_react_agent(
            llm=self.llm,
            tools=self.tools,
            prompt=self.prompt
        )

        self.agent_executor = AgentExecutor(
            agent=self.agent,
            tools=self.tools,
            verbose=False,
            handle_parsing_errors=True,
            callbacks=[RichCallbackHandler()] if verbose else [],
        )

        self.memory = ConversationBufferMemory(return_messages=True)

    def retrieve(self, query: str, verbose: bool = False) -> str:
        """
        Invokes the agent to process the query and provides the final answer.
        """
        cm = (contextlib.nullcontext() if self.verbose
              else Console().status('', spinner='dots'))
        with cm:
            try:
                response = self.agent_executor.invoke({
                    "question": query,
                    "chat_history": self.memory.buffer_as_messages,
                })["output"]

                self.memory.save_context(
                    {"question": query}, {"output": response})

                return response
            except Exception as e:
                # Handle potential errors during execution
                return f"An error occurred: {e}"



from .base_agent import BaseAgent
from .hf_agent import HFDocAgent
from .local_agent import LocalDocAgent
from .local_react_agent import LocalDocReActAgent
from .pydantic_ai_agent import PydanticAIDocAgent


__all__ = ["BaseAgent", "HFDocAgent", "LocalDocAgent",
           "LocalDocReActAgent", "PydanticAIDocAgent"]
